/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.backend;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import com.redhat.thermostat.backend.adverse.AdverseBackend;
import com.redhat.thermostat.backend.adverse.AdverseBackendActivation;
import com.redhat.thermostat.backend.adverse.AdverseBackendDeactivation;
import org.junit.Test;

public class BackendActivatorTest {

    @Test(expected = IllegalArgumentException.class)
    public void testDoesNotAcceptNegativeArguments() {
        new BackendActivator(-5, -2);
    }

    @Test
    public void testActivateBackend() {
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(true);
        Runnable postActivateRunnable = mock(Runnable.class);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(3, scheduledExecutorService, scheduledExecutorService,
                                                                 scheduledExecutorService, executorService, new CountDownLatch(1));

        assertEquals(0, backendActivator.getNumberOfBackendsActive());
        backendActivator.activateBackend(backend, postActivateRunnable);
        assertEquals(1, backendActivator.getNumberOfBackendsActive());

        verify(backend, times(1)).activate();
        verify(postActivateRunnable, times(1)).run();
    }

    @Test
    public void testDeactivateBackend() {
        Backend backend = mock(Backend.class);
        when(backend.deactivate()).thenReturn(true);
        Runnable postDeactivateRunnable = mock(Runnable.class);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(3, scheduledExecutorService, scheduledExecutorService,
                                                                 scheduledExecutorService, executorService, new CountDownLatch(1));
        backendActivator.deactivateBackend(backend, postDeactivateRunnable);

        verify(backend, times(1)).deactivate();
        verify(postDeactivateRunnable, times(1)).run();
    }

    @Test
    public void testActivateBackendNotTrackedOrPostActivated() {
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(false);
        Runnable postActivateRunnable = mock(Runnable.class);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(3, scheduledExecutorService, scheduledExecutorService,
                scheduledExecutorService, executorService, new CountDownLatch(1));
        backendActivator.activateBackend(backend, postActivateRunnable);

        verify(backend, times(1)).activate();
        verify(postActivateRunnable, times(0)).run();
        assertEquals(1, backendActivator.getNumberOfBackendsActive());
    }

    @Test
    public void testDeactivateBackendNotTrackedOrPostDeactivated() {
        Backend backend = mock(Backend.class);
        when(backend.deactivate()).thenReturn(false);
        Runnable postDeactivateRunnable = mock(Runnable.class);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(3, scheduledExecutorService, scheduledExecutorService,
                scheduledExecutorService, executorService, new CountDownLatch(1));
        backendActivator.deactivateBackend(backend, postDeactivateRunnable);

        verify(backend, times(1)).deactivate();
        verify(postDeactivateRunnable, times(0)).run();
        assertEquals(0, backendActivator.getNumberOfBackendsActive());
    }

    @Test
    public void testActivateAndDeactivate() {
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(true);
        when(backend.deactivate()).thenReturn(true);
        Runnable postActivateRunnable = mock(Runnable.class);
        Runnable postDeactivateRunnable = mock(Runnable.class);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(3, scheduledExecutorService, scheduledExecutorService,
                                                                 scheduledExecutorService, executorService, new CountDownLatch(1));

        assertEquals(0, backendActivator.getNumberOfBackendsActive());
        backendActivator.activateBackend(backend, postActivateRunnable);
        assertEquals(1, backendActivator.getNumberOfBackendsActive());
        backendActivator.deactivateBackend(backend, postDeactivateRunnable);
        assertEquals(0, backendActivator.getNumberOfBackendsActive());

        verify(backend, times(1)).activate();
        verify(backend, times(1)).deactivate();
        verify(postActivateRunnable, times(1)).run();
        verify(postDeactivateRunnable, times(1)).run();
    }

    @Test(timeout = 3000)
    public void testAdverseActivation() {
        AdverseBackend backend = new AdverseBackendActivation();
        Runnable postDeactivateRunnable = mock(Runnable.class);

        BackendActivator backendActivator = new BackendActivator(2, 1);
        backendActivator.activateBackend(backend, postDeactivateRunnable);
        backend.waitUntilInterrupted();
    }

    @Test(timeout = 3000)
    public void testAdverseDeactivation() {
        AdverseBackend backend = new AdverseBackendDeactivation();
        Runnable postDeactivateRunnable = mock(Runnable.class);

        BackendActivator backendActivator = new BackendActivator(2, 1);
        backendActivator.deactivateBackend(backend, postDeactivateRunnable);
        backend.waitUntilInterrupted();
    }

    @Test(timeout = 1000)
    public void testCountsDownBackend() {
        CountDownLatch countDownLatch = mock(CountDownLatch.class);
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(true);
        when(backend.deactivate()).thenReturn(true);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(1, scheduledExecutorService, scheduledExecutorService,
                                                                 scheduledExecutorService, executorService, countDownLatch);

        verify(countDownLatch, times(0)).countDown();
        backendActivator.activateBackend(backend, mock(Runnable.class));
        verify(countDownLatch, times(0)).countDown();

        backendActivator.deactivateBackend(backend, mock(Runnable.class));
        verify(countDownLatch, times(1)).countDown();
    }

    @Test(timeout = 1000)
    public void testWaitsForDeactivation() {
        CountDownLatch countDownLatch = new CountDownLatch(1);
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(true);
        when(backend.deactivate()).thenReturn(true);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(1, scheduledExecutorService, scheduledExecutorService,
                                                                 scheduledExecutorService, executorService, countDownLatch);

        backendActivator.activateBackend(backend, mock(Runnable.class));
        backendActivator.deactivateBackend(backend, mock(Runnable.class));

        // This should guarantee we overshoot the timeout if the implementation
        // is broken (by not waiting on the latch, or not counting down).
        backendActivator.waitForAllBackendsToDeactivate(60000);
    }

    @Test
    public void testCannotAddActivationAfterShutdown() {
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(true);
        when(backend.deactivate()).thenReturn(true);
        Runnable runnable = mock(Runnable.class);
        BackendActivator backendActivator = new BackendActivator(2, 10);

        backendActivator.shutdown();

        assertEquals(0, backendActivator.getNumberOfBackendsActive());
        backendActivator.activateBackend(backend, runnable);
        assertEquals(0, backendActivator.getNumberOfBackendsActive());
    }

    @Test
    public void testCannotAddDeactivationAfterShutdown() {
        Backend backend = mock(Backend.class);
        when(backend.activate()).thenReturn(true);
        when(backend.deactivate()).thenReturn(true);
        Runnable runnable = mock(Runnable.class);
        CountDownLatch countDownLatch = new CountDownLatch(1);
        ScheduledExecutorService scheduledExecutorService = new InstantScheduledThreadPoolExecutor();
        ExecutorService executorService = new InstantExecutorService();
        BackendActivator backendActivator = new BackendActivator(1, scheduledExecutorService, scheduledExecutorService,
                                                                 scheduledExecutorService, executorService, countDownLatch);

        backendActivator.activateBackend(backend, runnable);
        assertEquals(1, backendActivator.getNumberOfBackendsActive());

        backendActivator.shutdown();

        backendActivator.deactivateBackend(backend, runnable);
        assertEquals(1, backendActivator.getNumberOfBackendsActive());
    }

    private static class InstantScheduledThreadPoolExecutor extends ScheduledThreadPoolExecutor {
        private InstantScheduledThreadPoolExecutor() {
            super(1);
        }

        @Override
        public ScheduledFuture<?> schedule(Runnable runnable, long timeout, TimeUnit timeoutTimeUnit) {
            ScheduledFuture scheduledFutureMock = mock(ScheduledFuture.class);
            when(scheduledFutureMock.isDone()).thenReturn(true);
            runnable.run();
            return scheduledFutureMock;
        }
    }

    private static class InstantExecutorService extends ThreadPoolExecutor {
        private InstantExecutorService() {
            super(1, 1, 0L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue());
        }

        @Override
        public Future<?> submit(Runnable runnable) {
            Future futureMock = mock(Future.class);
            when(futureMock.isDone()).thenReturn(true);
            runnable.run();
            return futureMock;
        }
    }
}
